
/*
 * Supplemental material for article
 *
 * ``A review of polyhedral intersection detection and new techniques''
 * Author : Samuel Hornus
 * Copyright © Inria, 2015
 * Licence: Creative Commons CC BY-ND 3.0 license available online at
 * http://creativecommons.org/licenses/by-nd/3.0/
 */
 
/*
 * This file includes two implementations of a partial frustum culling of a triangle.
 * These functions do NOT test the triangle against the frustum's four planes
 * (this should be done separately), but compute--and test the triangle
 * against--planes tangent to the frustum with normal vector in N_{AB} using
 * the notation from the submitted paper.
 */

/*
 * FUNCTION noTriangleEdgeSeparatesFrustumNoSIMD
 *
 * Returns True if NO triangle edge generates a plane that separate the frustum
 * from the triangle. False otherwise.
 *
 * Parameters:
 *   - prim: is the triangle number inside the Indexed Triangle Set.
 *   - frustumOrigin is the apex of the frustum : camera center or point light source. 
 *   - frustumPlanes are the plane equations of the four planes bounding the frustum.
 *     Two consecutive planes MUST contain a common corner ray (so the 4 planes
 *     form a cycle of planes around the frustum, for example: top, right, bottom, left).
 */
bool
IndexedTriangleSet::
noTriangleEdgeSeparatesFrustumNoSIMD(
    const int prim,
    const Vec3f & frustumOrigin,
    const Vec4f frustumPlanes[4]) const
{
  auto edgeMakesNoSeparation = [&](
      const Vec3f & edge,
      const Vec3f & edgeVertex,
      const Vec3f & nextEdge,
      const Vec3f & otherVertex)
  {
    const float nv[4] = {
      frustumPlanes[0] | edge, // The operator | is a dot product
      frustumPlanes[1] | edge,
      frustumPlanes[2] | edge,
      frustumPlanes[3] | edge
    };

    auto testCorner = [&](const int i0, const int i1) {
      if( nv[i0] * nv[i1] >= 0.0f ) {
        // corner ray does not form a silhouette edge (of the frustum)
        // w.r.t. the direction given by 'edge' ==> No Separation Here.
        return true;
      }
      // Use linear interpolation to get normal vector to plane orthogonal to
      // both 'edge' and corner ray:
      const Vec3f sepPlaneNormal = (nv[i0] > 0.0f) ?
          (nv[i0] * frustumPlanes[i1].toVec3() - nv[i1] * frustumPlanes[i0].toVec3())
        : (nv[i1] * frustumPlanes[i0].toVec3() - nv[i0] * frustumPlanes[i1].toVec3());

      /* UN-comment the line below to truly compute the exact Gauss map overlay.
       * Keeping the line commented has the effect of considering the full
       * great-circle if the dual-arcs of the triangle, as is done in the SSE
       * implementation below.
       */
      //if( (sepPlaneNormal | nextEdge) <= 0.0f ) return true;
      // here, we have a plane on the minkoswki difference

      const float frustumOrgPos = sepPlaneNormal | frustumOrigin;
      return
        ( (sepPlaneNormal | edgeVertex)  <= frustumOrgPos ) ||
        ( (sepPlaneNormal | otherVertex) <= frustumOrgPos );
    };

    return testCorner(0,1) && testCorner(1,2) && testCorner(2,3) && testCorner(3,0);
  };

  const Vec3ui & vIdx = (*this)[prim]; // get the triangle's vertex indices
  // Positive triangle : v0 -> v1 -> v2
  // Negative triangle : -v0 -> -v1 -> -v2 -> -v0
  const Vec3f e0 = vertices[vIdx(0)] - vertices[vIdx(1)];
  const Vec3f e1 = vertices[vIdx(1)] - vertices[vIdx(2)];
  const Vec3f e2 = vertices[vIdx(2)] - vertices[vIdx(0)];
  return
    edgeMakesNoSeparation(e0, vertices[vIdx(0)], e2, vertices[vIdx(2)]) &&
    edgeMakesNoSeparation(e1, vertices[vIdx(1)], e0, vertices[vIdx(0)]) &&
    edgeMakesNoSeparation(e2, vertices[vIdx(2)], e1, vertices[vIdx(1)]);
}

/*
 * FUNCTION noTriangleEdgeSeparatesFrustum
 *
 * SSE implementation of the function above.
 *
 * This function tests if there exists one edge of the triangle that supports
 * a plane separating the triangle from the frustum bounded by four planes.
 * It tests the triangle's 3 edges consecutively. For each edge, the four
 * "corner rays" of the frustum are tested simultaneously with SSE.
 */
bool
IndexedTriangleSet::
noTriangleEdgeSeparatesFrustum(
    const int prim,
    const Vec3f & frustumOrigin,
    const Vec4f frustumPlanes[4]) const
{
  // TODO: it might be faster to perform four _mm_load_ps followed by partial matrix transpose.
  const __m128 frustumNormals_x = _mm_set_ps(frustumPlanes[0].x(),frustumPlanes[1].x(),frustumPlanes[2].x(),frustumPlanes[3].x());
  const __m128 frustumNormals_y = _mm_set_ps(frustumPlanes[0].y(),frustumPlanes[1].y(),frustumPlanes[2].y(),frustumPlanes[3].y());
  const __m128 frustumNormals_z = _mm_set_ps(frustumPlanes[0].z(),frustumPlanes[1].z(),frustumPlanes[2].z(),frustumPlanes[3].z());

  const Vec3ui & vIdx = (*this)[prim];
  const Vec3f & p0 = vertices[vIdx(0)];
  const Vec3f & p1 = vertices[vIdx(1)];
  const Vec3f & p2 = vertices[vIdx(2)];
  // Positive triangle : v0 -> v1 -> v2
  // Negative triangle : -v0 -> -v1 -> -v2 -> -v0
  const __m128 zero = _mm_setzero_ps();

  const int rotationMask = _MM_SHUFFLE(0,3,2,1);

  auto edgeMakesNoSeparation = [&](
      const Vec3f & p0,
      const Vec3f & p1,
      const Vec3f & p2)
  {
    const __m128 nv =
      _mm_add_ps( // four dot products : 4 frustum normals VS one triangle edge
          _mm_add_ps(
            _mm_mul_ps(frustumNormals_x, _mm_set1_ps(p0.x() - p1.x())),
            _mm_mul_ps(frustumNormals_y, _mm_set1_ps(p0.y() - p1.y()))),
          _mm_mul_ps(frustumNormals_z, _mm_set1_ps(p0.z() - p1.z())));

    const __m128 nvShifted = _mm_shuffle_ps(nv, nv, rotationMask);
    const __m128 bothSides = _mm_cmplt_ps(_mm_mul_ps(nv, nvShifted), zero);
    const __m128 multiplier = _mm_blendv_ps(_mm_set1_ps(1.0f), _mm_set1_ps(-1.0f), nv);

    // 3 * 4 linear interpolations (normalizing weights to one is NOT done because it is useless.)
    const __m128 sepNormals_x = _mm_mul_ps(multiplier,
        _mm_sub_ps(
          _mm_mul_ps(nv, _mm_shuffle_ps(frustumNormals_x, frustumNormals_x, rotationMask)),
          _mm_mul_ps(nvShifted, frustumNormals_x)));
    const __m128 sepNormals_y = _mm_mul_ps(multiplier,
        _mm_sub_ps(
          _mm_mul_ps(nv, _mm_shuffle_ps(frustumNormals_y, frustumNormals_y, rotationMask)),
          _mm_mul_ps(nvShifted, frustumNormals_y)));
    const __m128 sepNormals_z = _mm_mul_ps(multiplier,
        _mm_sub_ps(
          _mm_mul_ps(nv, _mm_shuffle_ps(frustumNormals_z, frustumNormals_z, rotationMask)),
          _mm_mul_ps(nvShifted, frustumNormals_z)));

    const __m128 frustumOrgPos = // 4 dot products
      _mm_add_ps(
          _mm_add_ps(
            _mm_mul_ps(sepNormals_x, _mm_set1_ps(frustumOrigin.x())),
            _mm_mul_ps(sepNormals_y, _mm_set1_ps(frustumOrigin.y()))),
          _mm_mul_ps(sepNormals_z, _mm_set1_ps(frustumOrigin.z())));
    const __m128 edgePos = // 4 dot products
      _mm_add_ps(
          _mm_add_ps(
            _mm_mul_ps(sepNormals_x, _mm_set1_ps(p0.x())),
            _mm_mul_ps(sepNormals_y, _mm_set1_ps(p0.y()))),
          _mm_mul_ps(sepNormals_z, _mm_set1_ps(p0.z())));
    const __m128 separatesEdge = _mm_and_ps(bothSides, _mm_cmplt_ps(frustumOrgPos, edgePos));
    const __m128 otherPos = // 4 dot products
      _mm_add_ps(
          _mm_add_ps(
            _mm_mul_ps(sepNormals_x, _mm_set1_ps(p2.x())),
            _mm_mul_ps(sepNormals_y, _mm_set1_ps(p2.y()))),
          _mm_mul_ps(sepNormals_z, _mm_set1_ps(p2.z())));
    const __m128 separates = _mm_and_ps(separatesEdge, _mm_cmplt_ps(frustumOrgPos, otherPos));
    return ( 0 == _mm_movemask_ps(separates) );
  };
  return
    edgeMakesNoSeparation(p0, p1, p2) &&
    edgeMakesNoSeparation(p1, p2, p0) &&
    edgeMakesNoSeparation(p2, p0, p1);
}
